from __future__ import absolute_import
input_name = '../examples/diffusion/darcy_flow_multicomp.py'
output_name = 'test_darcy_flow_multicomp.vtk'

from tests_basic import TestInput
class Test(TestInput):
    pass
