from __future__ import absolute_import
input_name = '../examples/diffusion/poisson_field_dependent_material.py'
output_name_trunk = 'test_poisson_field_dependent_material'

from tests_basic import TestInputEvolutionary

class Test(TestInputEvolutionary):
    pass
