# c: 06.02.2008, r: 06.02.2008
from __future__ import absolute_import
input_name = '../examples/diffusion/time_poisson.py'
output_name_trunk = 'test_time_poisson'

from tests_basic import TestInputEvolutionary

class Test( TestInputEvolutionary ):
    pass
