from __future__ import absolute_import
input_name = '../examples/multi_physics/biot_short_syntax.py'
output_name = 'test_biot_short_syntax.vtk'

from tests_basic import TestInput

class Test(TestInput):
    pass
