from __future__ import absolute_import
input_name = '../examples/linear_elasticity/prestress_fibres.py'
output_name = 'test_prestress_fibres.vtk'

from tests_basic import TestInput

class Test(TestInput):
    pass
